@@ -1,21 +1,22 @@
 # -*- coding: utf-8 -*-
-"""
-Zegami Ltd.
-Apache 2.0
-"""
+# Copyright 2021 Zegami Ltd
 
+"""collection functionality."""
+
+from concurrent.futures import as_completed, ThreadPoolExecutor
 from io import BytesIO
+import json
 import os
+from time import time
+
 import pandas as pd
 from PIL import Image
-from time import time
-from concurrent.futures import ThreadPoolExecutor, as_completed
-import json
 
 from .source import Source
 
 
 class Collection():
+
     @staticmethod
     def _construct_collection(client, workspace, collection_dict, allow_caching=True):
         """Use this to instantiate Collection instances.
@@ -47,12 +48,10 @@ class Collection():
         self.allow_caching = allow_caching
         self._cached_rows = None
         self._cached_image_meta_lookup = None
-        self._cached_tags = None
 
     def clear_cache(self):
         self._cached_rows = None
         self._cached_image_meta_lookup = None
-        self._cached_tags = None
 
     @property
     def client():
@@ -100,7 +99,7 @@ class Collection():
     def _upload_dataset_id():
         pass
 
-    @_dataset_id.getter  # noqa: F811
+    @_upload_dataset_id.getter
     def _upload_dataset_id(self) -> str:
         self._check_data()
         assert 'upload_dataset_id' in self._data.keys(),\
@@ -151,7 +150,8 @@ class Collection():
     def sources(self):
         if self.version < 2:
             print('{} is an old-style collection and does not support multiple image sources'.format(self.name))
-            return []
+            #  return a constructed pseudo source object in order to provide a consistant interface
+            return [Source(self, self._data)]
 
     def show_sources(self):
         print('{} is an old-style collection and does not support multiple image sources'.format(self.name))
@@ -196,12 +196,17 @@ class Collection():
 
     @tags.getter
     def tags(self):
-        if self.allow_caching and self._cached_tags is not None:
-            return self._cached_tags
-        tags = self._get_tag_indices()
-        if self.allow_caching:
-            self._cached_tags = tags
-        return tags
+        return self._get_tag_indices()
+
+    @property
+    def status():
+        pass
+
+    @status.getter
+    def status(self):
+        details_url = f'{self.client.HOME}/{self.client.API_0}/project/{self.workspace_id}/collections/{self.id}'
+        resp = self.client._auth_get(details_url)
+        return resp['collection']['status']
 
     def get_rows_by_filter(self, filters):
         """Gets rows of metadata in a collection by a flexible filter.
@@ -235,17 +240,21 @@ class Collection():
         assert type(tag_names) == list,\
             'Expected tag_names to be a list, not a {}'.format(type(tag_names))
 
-        row_indicies = set()
+        row_indices = set()
         for tag in tag_names:
             if tag in self.tags.keys():
-                row_indicies.update(self.tags[tag])
-        rows = self.rows.iloc[list(row_indicies)]
+                row_indices.update(self.tags[tag])
+        rows = self.rows.iloc[list(row_indices)]
         return rows
 
-    def get_image_urls(self, rows, source=0):
+    def get_image_urls(self, rows, source=0, generate_signed_urls=False):
         """Converts rows into their corresponding image URLs.
 
-        You can use these URLs with download_image()/download_image_batch().
+        If generate_signed_urls is false the URLs require a token to download
+        These urls can be passed to download_image()/download_image_batch().
+
+        If generate_signed_urls is true the urls can be used to fetch the images directly
+        from blob storage, using a temporary access signature.
         """
         # Turn the provided 'rows' into a list of ints
         if type(rows) == pd.DataFrame:
@@ -264,18 +273,44 @@ class Collection():
 
         # Convert these into URLs
         c = self.client
-        return ['{}/{}/project/{}/imagesets/{}/images/{}/data'.format(
-            c.HOME, c.API_0, self.workspace_id, self._get_imageset_id(source),
-            i) for i in imageset_indices]
+        if not generate_signed_urls:
+            return ['{}/{}/project/{}/imagesets/{}/images/{}/data'.format(
+                c.HOME, c.API_0, self.workspace_id, self._get_imageset_id(source),
+                i) for i in imageset_indices]
+        else:
+            get_signed_urls = ['{}/{}/project/{}/imagesets/{}/images/{}/signed_route'.format(
+                c.HOME, c.API_0, self.workspace_id, self._get_imageset_id(source),
+                i
+            ) for i in imageset_indices]
+            signed_route_urls = []
+            for url in get_signed_urls:
+                response = c._auth_get(url)
+                signed_route_urls.append(response['url'])
+            return signed_route_urls
+
+    def download_annotation(self, annotation_id):
+        """
+        Converts an annotation_id into downloaded annotation data.
+
+        This will vary in content depending on the annotation type and
+        format.
+        """
+        zc = self.client
+        url = '{}/{}/project/{}/annotations/{}'.format(
+            zc.HOME, zc.API_1, self.workspace.id, annotation_id)
+        return zc._auth_get(url)
 
     def replace_data(self, data):
-        """Replaces the data in the collection.
-
-        The provided input should be a pandas dataframe or a local csv/json/tsv/txt/xlsx/xls file.
-        If a xlsx/xls file is used only data from the default sheet will be fetched.
         """
-        upload_data = ''
-        name = ''
+        Replaces the data in the collection.
+
+        The provided input should be a pandas dataframe or a local
+        csv/json/tsv/txt/xlsx/xls file. If a xlsx/xls file is used only data
+        from the default sheet will be fetched. The rows take time to be
+        updated, hence checking the status of the collection with coll.status,
+        might be helpful if you need to ensure that you're using the updated
+        data rows.
+        """
         if type(data) == pd.DataFrame:
             tsv = data.to_csv(sep='\t', index=False)
             upload_data = bytes(tsv, 'utf-8')
@@ -296,7 +331,12 @@ class Collection():
         mime_type = 'application/octet-stream'
 
         # create blob storage and upload to it
-        url, blob_id = zeg_client._obtain_signed_blob_storage_url(self.workspace_id)
+        urls, id_set = zeg_client._obtain_signed_blob_storage_urls(
+            self.workspace_id,
+            blob_path=f'datasets/{self._upload_dataset_id}'
+        )
+        blob_id = id_set['ids'][0]
+        url = urls[blob_id]
         zeg_client._upload_to_signed_blob_storage_url(upload_data, url, mime_type)
 
         # update the upload dataset details
@@ -306,10 +346,48 @@ class Collection():
 
         # returning response is true as otherwise it will try to return json but this response is empty
         zeg_client._auth_put(upload_dataset_url, body=None, return_response=True, json=current_dataset)
-        print(f'Dataset [id: {self._dataset_id}] updated successfully.')
+
+        self._cached_rows = None
+
+    def save_image(self, url, target_folder_path='./', filename='image', extension='png'):
+        """
+        Downloads an image and saves to disk
+
+        For input, see Collection.get_image_urls().
+        """
+        if not os.path.exists(target_folder_path):
+            os.makedirs(target_folder_path)
+
+        r = self.client._auth_get(url, return_response=True, stream=True)
+        with open(target_folder_path + '/' + filename + '.' + extension, 'wb') as f:
+            f.write(r.content)
+
+    def save_image_batch(self, urls, target_folder_path='./', extension='png', max_workers=50, show_time_taken=True):
+        """
+        Downloads a batch of images and saves to disk.
+
+        Filenames are the imageset index followed by the specified extension.
+        For input, see Collection.get_image_urls().
+        """
+        def save_single(index, url):
+            self.save_image(url, target_folder_path, filename=str(index), extension=extension)
+            return index
+
+        t = time()
+        with ThreadPoolExecutor(max_workers=max_workers) as ex:
+            futures = [ex.submit(save_single, i, u) for i, u in enumerate(urls)]
+
+            images = {}
+            for f in as_completed(futures):
+                i = f.result()
+                images[i] = 1
+
+                if show_time_taken:
+                    print('\nDownloaded {} images in {:.2f} seconds.'.format(len(images), time() - t))
 
     def download_image(self, url):
-        """Downloads an image into memory as a PIL.Image.
+        """
+        Downloads an image into memory as a PIL.Image.
 
         For input, see Collection.get_image_urls().
         """
@@ -318,11 +396,13 @@ class Collection():
         return Image.open(r.raw)
 
     def download_image_batch(self, urls, max_workers=50, show_time_taken=True):
-        """Downloads multiple images into memory (each as a PIL.Image) concurrently.
+        """
+        Downloads multiple images into memory (each as a PIL.Image) concurrently.
 
         Please be aware that these images are being downloaded into memory,
         if you download a huge collection of images you may eat up your
-        RAM! """
+        RAM!
+        """
         def download_single(index, url):
             return (index, self.download_image(url))
 
@@ -335,9 +415,9 @@ class Collection():
                 i, img = f.result()
                 images[i] = img
 
-        if show_time_taken:
-            print('\nDownloaded {} images in {:.2f} seconds.'.format(
-                len(images), time() - t))
+                if show_time_taken:
+                    print('\nDownloaded {} images in {:.2f} seconds.'.format(
+                        len(images), time() - t))
 
         # Results are a randomly ordered dictionary of results, so reorder them
         ordered = []
@@ -345,8 +425,23 @@ class Collection():
             ordered.append(images[i])
         return ordered
 
+    def delete_images_with_tag(self, tag='delete'):
+        """Delete all the images in the collection with the tag 'delete'.s."""
+        row_indices = set()
+        if tag in self.tags.keys():
+            row_indices.update(self.tags[tag])
+            lookup = self._get_image_meta_lookup()
+            imageset_indices = [lookup[int(i)] for i in row_indices]
+            c = self.client
+            urls = ['{}/{}/project/{}/imagesets/{}/images/{}'.format(
+                c.HOME, c.API_0, self.workspace_id, self._get_imageset_id(),
+                i) for i in imageset_indices]
+            for url in urls:
+                c._auth_delete(url)
+            print(f'\nDeleted {len(urls)} images')
+
     def _get_tag_indices(self):
-        """Returns collection tags indicies."""
+        """Returns collection tags indices."""
         c = self.client
         url = '{}/{}/project/{}/collections/{}/tags'.format(
             c.HOME, c.API_1, self.workspace_id, self.id)
@@ -354,7 +449,7 @@ class Collection():
         return self._parse_tags(response['tagRecords'])
 
     def _parse_tags(self, tag_records):
-        """Parses tag indicies into a list of tags, each with an list of indicies."""
+        """Parses tag indices into a list of tags, each with an list of indices."""
         tags = {}
         for record in tag_records:
             if record['tag'] not in tags.keys():
@@ -362,19 +457,20 @@ class Collection():
             tags[record['tag']].append(record['key'])
         return tags
 
-    def get_annotations(self, source=None) -> list:
-        """Returns all annotations attached to the collection."""
-        if source is not None:
-            self._source_warning()
+    def get_annotations(self, anno_type='mask') -> dict:
+        """
+        Returns one type of annotations attached to the collection.
 
+        Default as mask annotations.
+        """
         c = self.client
-        url = '{}/{}/project/{}/annotations/collection/{}'.format(
-            c.HOME, c.API_1, self.workspace_id, self.id)
+        url = '{}/{}/project/{}/annotations/collection/{}?type={}'.format(
+            c.HOME, c.API_1, self.workspace_id, self.id, anno_type)
 
         return c._auth_get(url)
 
-    def get_annotations_for_image(self, row_index, source=None) -> list:
-        """Returns all annotations for a single item in the collection."""
+    def get_annotations_for_image(self, row_index, source=None, anno_type='mask') -> list:
+        """Returns one type of annotations for a single item in the collection. Default as mask annotations."""
         if source is not None:
             self._source_warning()
 
@@ -385,20 +481,20 @@ class Collection():
         lookup = self._get_image_meta_lookup()
         imageset_index = lookup[row_index]
 
-        url = '{}/{}/project/{}/annotations/imageset/{}/images/{}'.format(
-            c.HOME, c.API_1, self.workspace_id, self._get_imageset_id(), imageset_index)
+        url = '{}/{}/project/{}/annotations/imageset/{}/images/{}?type={}'\
+            .format(c.HOME, c.API_1, self.workspace_id, self._get_imageset_id(), imageset_index, anno_type)
 
         return c._auth_get(url)
 
-    def upload_annotation(
-            self, uploadable, row_index=None, image_index=None, source=None,
-            author=None, debug=False):
-        """Uploads an annotation to Zegami.
+    def upload_annotation(self, uploadable, row_index=None, image_index=None, source=None, author=None, debug=False):
+        """
+        Uploads an annotation to Zegami.
 
-        Requires uploadable annotation data (see AnnotationClass.create_uploadable), the row index of
-        the image the annotation belongs to, and the source (if using a
-        multi-image-source collection). If no source is provided, it will be
-        uploaded to the first source.
+        Requires uploadable annotation data (see
+        AnnotationClass.create_uploadable()), the row index of the image the
+        annotation belongs to, and the source (if using a multi-image-source
+        collection). If no source is provided, it will be uploaded to the
+        first source.
 
         Optionally provide an author, which for an inference result should
         probably some identifier for the model. If nothing is provided, the
@@ -418,44 +514,173 @@ class Collection():
                 'Must provide only one or row_index, image_index'
 
         assert type(uploadable) == dict,\
-            'Expected uploadable data to be a dict, not a {}'.format(type(uploadable))
+            'Expected uploadable data to be a dict, not a {}'\
+            .format(type(uploadable))
+
         assert 'type' in uploadable.keys(),\
             'Expected \'type\' key in uploadable: {}'.format(uploadable)
+
         assert 'annotation' in uploadable.keys(),\
             'Expected \'annotation\' key in uploadable: {}'.format(uploadable)
+
         assert type(imageset_id) == str,\
-            'Expected imageset_id to be a str, not {}'.format(type(imageset_id))
+            'Expected imageset_id to be a str, not {}'\
+            .format(type(imageset_id))
 
         # Get the class-specific data to upload
         payload = {
+            'author': author,
             'imageset_id': imageset_id,
             'image_index': int(image_index),
-            'author': author,
-            'class_id': int(uploadable['class_id']),
+            'annotation': uploadable['annotation'],
             'type': uploadable['type'],
             'format': uploadable['format'],
-            'annotation': uploadable['annotation'],
+            'class_id': str(int(uploadable['class_id'])),
         }
 
         # Check that there are no missing fields in the payload
         for k, v in payload.items():
-            assert v, 'Empty annotation uploadable data value for \'{}\''.format(k)
+            assert v is not None, 'Empty annotation uploadable data value '
+            'for \'{}\''.format(k)
 
         # Potentially print for debugging purposes
         if debug:
             print('\nupload_annotation payload:\n')
             for k, v in payload.items():
-                print('{} : {}'.format(k, v))
+                if k == 'annotation':
+                    print('- annotation:')
+                    for k2, v2 in payload['annotation'].items():
+                        print('\t- {} : {}'.format(k2, v2))
+                else:
+                    print('- {} : {}'.format(k, v))
             print('\nJSON:\n{}'.format(json.dumps(payload)))
 
         # POST
         c = self.client
-        url = '{}/{}/project/{}/annotations/'.format(
+        url = '{}/{}/project/{}/annotations'.format(
             c.HOME, c.API_1, self.workspace_id)
         r = c._auth_post(url, json.dumps(payload), return_response=True)
 
         return r
 
+    def delete_annotation(self, annotation_id):
+        """Delete an annotation by its ID. These are obtainable using the get_annotations...() methods."""
+        c = self.client
+        url = '{}/{}/project/{}/annotations/{}'\
+            .format(c.HOME, c.API_1, self.workspace_id, annotation_id)
+        payload = {
+            'author': c.email
+        }
+        r = c._auth_delete(url, data=json.dumps(payload))
+
+        return r
+
+    def delete_all_annotations(self, source=0):
+        """Deletes all annotations saved to the collection."""
+        # A list of sources of annotations
+        anno_sources = self.get_annotations()['sources']
+
+        c = 0
+        for i, source in enumerate(anno_sources):
+
+            # A list of annotation objects
+            annotations = source['annotations']
+            if len(annotations) == 0:
+                continue
+
+            print('Deleting {} annotations from source {}'
+                  .format(len(annotations), i))
+
+            for j, annotation in enumerate(annotations):
+                self.delete_annotation(annotation['id'])
+                print('\r{}/{}'.format(j + 1, len(annotations)), end='',
+                      flush=True)
+                c += 1
+            print('')
+
+        print('\nDeleted {} annotations from collection "{}"'.format(
+            c, self.name))
+
+    @property
+    def userdata():
+        pass
+
+    @userdata.getter
+    def userdata(self):
+        c = self.client
+        url = '{}/{}/project/{}/collections/{}'.format(
+            c.HOME, c.API_0, self.workspace_id, self.id)
+        data = c._auth_get(url)['collection']
+        userdata = data['userdata'] if 'userdata' in data.keys() else None
+        return userdata
+
+    @property
+    def classes():
+        pass
+
+    @classes.getter
+    def classes(self) -> list:
+        """
+        Property for the class configuration of the collection.
+
+        Used in an annotation workflow to tell Zegami how to treat defined classes.
+
+        To set new classes, provide a list of class dictionaries of shape:
+
+        collection.classes = [
+            {
+                'color' : '#32a852',    # A hex color for the class
+                'name'  : 'Dog',        # A human-readable identifier for the class
+                'id'    : 0             # The unique integer class ID
+            },
+            {
+                'color' : '#43f821',    # A hex color for the class
+                'name'  : 'Cat',        # A human-readable identifier for the class
+                'id'    : 1             # The unique integer class ID
+            }
+        ]
+        """
+        u = self.userdata
+        return list(u['classes'].values()) if u is not None and 'classes' in u.keys() else []
+
+    @classes.setter
+    def classes(self, classes):  # noqa: C901
+        # Check for a valid classes list
+        if type(classes) != list:
+            raise TypeError('Expected \'classes\' to be a list, not {}'.format(type(classes)))
+
+        payload = {
+            'classes': {}
+        }
+
+        for d in classes:
+            # Check for a sensible class dict
+            if type(d) != dict:
+                raise TypeError('Expected \'classes\' entry to be a dict, not {}'.format(type(d)))
+            if len(d.keys()) != 3:
+                raise ValueError('Expected classes dict to have 3 keys, not {} ({})'.format(len(d.keys()), d))
+            for k in ['color', 'name', 'id']:
+                if k not in d.keys():
+                    raise ValueError('Unexpected class key: {}. Keys must be '
+                                     'color | name | id.'.format(k))
+
+            # Format as the expected payload
+            payload['classes'][d['id']] = {
+                'color': str(d['color']),
+                'name': str(d['name']),
+                'id': str(int(d['id']))
+            }
+
+        # POST
+        c = self.client
+        url = '{}/{}/project/{}/collections/{}/userdata'.format(
+            c.HOME, c.API_0, self.workspace_id, self.id)
+        c._auth_post(url, json.dumps(payload))
+
+        print('New classes set:')
+        for d in self.classes:
+            print(d)
+
     def _check_data(self) -> None:
         assert self._data, 'Collection had no self._data set'
         assert type(self._data) == dict,\
@@ -511,10 +736,11 @@ class Collection():
         return self._data['total_data_items']
 
     def __repr__(self) -> str:
-        return "Collection id={} name={}".format(self.id, self.name)
+        return "<Collection id={} name={}>".format(self.id, self.name)
 
 
 class CollectionV2(Collection):
+
     @property
     def sources():
         pass
@@ -527,32 +753,38 @@ class CollectionV2(Collection):
         return [Source(self, s) for s in self._data['image_sources']]
 
     def show_sources(self):
-        """lists the IDs and names of all sources in the collection."""
+        """Lists the IDs and names of all sources in the collection."""
         ss = self.sources
         print('\nImage sources ({}):'.format(len(ss)))
         for s in ss:
-            print('{} : {}'.format(s._imageset_id, s.name))
+            print('{} : {}'.format(s.imageset_id, s.name))
 
-    def get_annotations(self, source=0) -> list:
-        """Gets all annotations for a particular source of a collection."""
+    def get_annotations(self, source=0, anno_type='mask') -> list:
+        """
+        Gets one type of annotations for a particular source of a collection.
+
+        Default as mask annotations.
+        """
         c = self.client
         source = self._parse_source(source)
 
-        url = '{}/{}/project/{}/annotations/collection/{}/source/{}'.format(
-            c.HOME, c.API_1, self.workspace_id, self.id, source.id)
+        url = '{}/{}/project/{}/annotations/collection/{}/source/{}?type={}'.format(
+            c.HOME, c.API_1, self.workspace_id, self.id, source.id, anno_type)
         return c._auth_get(url)
 
-    def get_annotations_for_image(self, row_index, source=0) -> list:
-        """Returns all annotations for a single item in the collection."""
+    def get_annotations_for_image(self, row_index, source=0, anno_type='mask') -> list:
+        """
+        Returns one type of annotations for a single item in the collection.
+
+        Default as mask annotations.
+        """
         c = self.client
         source = self._parse_source(source)
-        assert type(row_index) == int and row_index >= 0,\
-            'Expected row_index to be a positive int, not {}'.format(row_index)
 
-        lookup = self._get_image_meta_lookup()
+        lookup = self._get_image_meta_lookup(source)
         imageset_index = lookup[row_index]
-        url = '{}/{}/project/{}/annotations/imageset/{}/images/{}'.format(
-            c.HOME, c.API_1, self.workspace_id, self._get_imageset_id(), imageset_index)
+        url = '{}/{}/project/{}/annotations/imageset/{}/images/{}?type={}'.format(
+            c.HOME, c.API_1, self.workspace_id, self._get_imageset_id(), imageset_index, anno_type)
 
         return c._auth_get(url)
 
@@ -560,7 +792,7 @@ class CollectionV2(Collection):
         """Source can be an int or a Source instance from this collection."""
         self._check_data()
         self._check_version()
-        return self._parse_source(source)._imageset_id
+        return self._parse_source(source).imageset_id
 
     def _get_image_meta_lookup(self, source=0) -> list:
         self._check_data()
@@ -590,3 +822,6 @@ class CollectionV2(Collection):
 
         raise Exception('Provided source was a Source instance, but didn\'t '
                         'belong to this collection ({})'.format(self.name))
+
+    def __repr__(self) -> str:
+        return "<Collection V2 id={} name={}>".format(self.id, self.name)
